@@ -14,13 +14,15 @@ import pkgutil
 import gnupg
 import base64
 import shutil
-logging.basicConfig(level=logging.INFO)
 
 import bootstrap_cfn.config as config
-from fabric.api import env, task, local, get, settings, sudo
+from fabric.api import env, execute, parallel, task, \
+    local, get, settings, sudo
 from fabric.contrib import files
+import fabric.decorators
+from fabric.exceptions import NetworkError
 from bootstrap_cfn.fab_tasks import _validate_fabric_env, \
-    get_stack_name, get_config, cfn_create, cfn_delete
+    get_stack_name, get_basic_config, cfn_create, cfn_delete
 
 from ec2 import EC2
 from bootstrap_salt.kms import KMS
@@ -29,6 +31,9 @@ from bootstrap_salt.config import MyConfigParser
 
 from .deploy_lib import github
 
+# Setup logging
+logging.basicConfig(level=logging.INFO)
+
 # GLOBAL VARIABLES
 env.aws = None
 TIMEOUT = 3600
@@ -45,6 +50,10 @@ env.stack = None
 # config of the auto scaling groups
 env.cloudformation_parser = MyConfigParser
 
+# Set the default bootstrap.sh location
+env.bootstrap_script_path = '/usr/local/bin'
+env.bootstrap_tmp_path = '/tmp'
+
 
 @task
 def aws(profile_name):
@@ -191,15 +200,15 @@ def wait_for_minions(timeout=600, interval=20):
 
     Args:
         timeout(int): time to wait for bootstrap to finish
-        interval(int): time to wait inbetween checks
+        interval(int): time to wait in-between checks
     """
     _validate_fabric_env()
     stack_name = get_stack_name()
     ec2 = get_connection(EC2)
-    print "Waiting for SSH on all instances..."
+    logging.info("Waiting for SSH on all instances...")
     ec2.wait_for_ssh(stack_name)
     fab_hosts = get_instance_ips()
-    print "Waiting for bootstrap script to finish on all instances..."
+    logging.info("Waiting for bootstrap script to finish on all instances...")
     utils.timeout(timeout, interval)(is_bootstrap_done)(fab_hosts)
 
 
@@ -207,7 +216,9 @@ def is_bootstrap_done(hosts):
     """
     Loops through a list of IPs and checks for the prescence of
     /tmp/bootstrap_done to ensure that the launch config has finished
-    executing.
+    executing. The call will also try to handle the case where we lose
+    ssh during the wait, which can happen if another action triggers
+    a reboot.
 
     Args:
         hosts(list): A list of IPs to check
@@ -215,17 +226,36 @@ def is_bootstrap_done(hosts):
     ret = []
     for host in hosts:
         env.host_string = '{0}@{1}'.format(env.user, host)
-        ret.append(files.exists('/tmp/bootstrap_done'))
+        target_file = '{}/bootstrap_done'.format(env.bootstrap_tmp_path)
+        try:
+            file_exists = files.exists(target_file)
+            ret.append(file_exists)
+            if file_exists:
+                logging.info("Salt bootstrap file {} on host {}..."
+                             .format(target_file, host))
+            else:
+                logging.info("Salt bootstrap file {} not found on host {}..."
+                             .format(target_file, host))
+        except NetworkError:
+            logging.warning("Could not connect to host {}, attempting to recover connection..."
+                            .format(host))
+            # Catch a network error and try again
+            stack_name = get_stack_name()
+            ec2 = get_connection(EC2)
+            ec2.wait_for_ssh(stack_name)
     return all(ret)
 
 
-def get_connection(klass):
+def get_connection(klass, optional=False):
     """
     Helper method to get connection to AWS
 
     Args:
         klass(class): The AWS class to setup the connection to.
     """
+    if optional and env.aws:
+        return klass(env.aws, env.aws_region)
+
     _validate_fabric_env()
     return klass(env.aws, env.aws_region)
 
@@ -342,7 +372,7 @@ def upload_salt():
 
 
 @task
-def encrypt_file(file_name, key_file="./salt.key.enc"):
+def encrypt_file(file_name, key_file="./salt.key.enc", kms_conn=None):
     """
     Encrypt a file using an encrypted KMS data key using GPG with an AES256
     cipher. Output file_name.gpg
@@ -352,7 +382,11 @@ def encrypt_file(file_name, key_file="./salt.key.enc"):
         key_file(string): path to encrypted key. Contents will be read and
             decrypted using KMS
     """
-    kms = get_connection(KMS)
+    if kms_conn is None:
+        kms = get_connection(KMS)
+    else:
+        kms = kms_conn
+
     if isinstance(key_file, basestring):
         key_file = open(key_file)
     key = kms.decrypt(key_file.read())['Plaintext']
@@ -383,13 +417,16 @@ def generate_ssh_key_pillar(force=False, strict=True):
     # strict: remove all users not found in github
     work_dir = os.path.dirname(env.real_fabfile)
 
-    config = get_config()
-    salt_cfg = config.data.get('salt', {})
+    # ssh key lookup only needs the basic configuration data in
+    # the local config.
+    basic_config = get_basic_config()
+    salt_cfg = basic_config.get('salt', {})
+    key_config = basic_config.get('github_users', {})
 
     local_pillar_dir = os.path.join(
         work_dir, salt_cfg.get('local_pillar_dir', 'pillar'))
     pillar_file = os.path.join(local_pillar_dir, env.environment, 'keys.sls')
-    key_config = config.data.get('github_users', {})
+
     ssh_key_data = github.get_keys(key_config)
 
     if os.path.exists(pillar_file) and not force:
@@ -449,3 +486,99 @@ def check_admins_exist():
                       % (', '.join(admins_dict.keys()),
                          host_ip)))
     return True
+
+
+@task
+def upgrade_packages(packages, fraction=None, restart=False):
+    """
+    Upgrade the packages specified, optionally rebooting afterwards
+
+    Args:
+        packages(list): List of packages to update. These can be a simple list
+            or with package versions, for example
+            '["package1", "package2"]' - get the latest versions of package1 and package2
+            '["package1", {"package2", "1.2.3"}]' - get the latest versions of package1,
+                and version 1.2.3 of package2.
+        fraction(float): The decimal fraction of minions to deploy to None means
+            all in one batch
+        restart(bool): False to not reboot after installation, True to reboot the instance.
+    """
+    for batch in get_ips_batch(fraction):
+        rup = fabric.decorators.hosts(batch)(run_upgrade_packages)
+        execute(rup, packages, fraction, restart)
+
+
+@parallel
+def run_upgrade_packages(packages, fraction=None, restart=False):
+    state = "pkg.install refresh=True only_upgrade=True pkgs='{}'".format(packages)
+    if fraction:
+        sudo('/usr/bin/salt-call {}'.format(state), shell=False)
+        if restart:
+            sudo('/usr/bin/salt-call system.reboot', shell=False)
+    else:
+        sudo('/usr/bin/salt-call {}'.format(state), shell=False)
+        if restart:
+            sudo('/usr/bin/salt-call system.reboot', shell=False)
+
+
+@task
+def update_users():
+    """
+    Setup a special fab environment to prepare for an all stack sync
+    user job.
+    """
+
+    try:
+        env.github_token = os.environ['GH_TOKEN']
+    except:
+        sys.exit("update_users requires a valid GH_TOKEN from github. Exiting...")
+
+    ec2_conn = get_connection(EC2, optional=True)
+    instances = ec2_conn.get_all_stack_ips()
+
+    # initialize lookup table
+    env.eght = dict()
+
+    if env.environment and env.application:
+        dec_func = fabric.decorators.hosts(get_instance_ips())(sync_users)
+    elif env.hosts:
+        dec_func = fabric.decorators.hosts(env.hosts)(sync_users)
+    else:
+        dec_func = fabric.decorators.hosts(instances.keys())(sync_users)
+
+    env.stack_names = instances
+    execute(dec_func)
+    local("rm /tmp/ght-*")
+
+
+@parallel(pool_size=2)
+def sync_users():
+    kms = get_connection(KMS, optional=True)
+
+    # Find stack id
+    stack = env.stack_names[env.host].split("-")
+    stack_name = "{0}-{1}".format(stack[0], stack[1])
+    filename = "/tmp/ght-{0}".format(stack_name)
+    s3_filename = "ght-{0}.gpg".format(env.stack_names[env.host])
+
+    if stack_name not in env.eght:
+        # Write token
+        fd = open(filename, "w")
+        fd.write(env.github_token)
+        fd.close()
+
+        # Encrypt out gh.token
+        key = StringIO.StringIO()
+        get(remote_path='/etc/salt.key.enc', local_path=key, use_sudo=True)
+        key.seek(0)
+        env.environment = stack[1]      # Setting the environment :(
+        encrypt_file(filename, key_file=key, kms_conn=kms)
+        key.close()
+        local("aws s3 --profile {0} cp {1}.gpg s3://{2}-salt/{3}".format(quote(env.aws),
+                                                                         quote(filename),
+                                                                         quote(env.stack_names[env.host]),
+                                                                         s3_filename))
+        env.eght[stack_name] = "env.stack_names[env.host].gpg"
+
+    sudo("salt-call saltutil.sync_modules")
+    sudo("salt-call github_user.addremove_users")
